### EXAMPLE 1: demonstrates how measurement error can lead to a type sign error in a covariate
### What kind of data invalidates fong + tyler?
### Even when you have a good predictor, if it's biased against a covariate you can get the wrong sign.
### Even when you include the proxy variable in the regression.
### But with some ground truth and multiple imputation, you can fix it.

library(argparser)
library(mecor)
library(ggplot2)
library(data.table)
library(filelock)
library(arrow)
library(Amelia)
library(Zelig)
library(predictionError)
options(amelia.parallel="no",
        amelia.ncpus=1)
setDTthreads(40)

source("simulation_base.R")

## SETUP:
### we want to estimate x -> y; x is MAR
### we have x -> k; k -> w; x -> w is used to predict x via the model w.
### A realistic scenario is that we have an NLP model predicting something like "racial harassment" in social media comments
### The labels x are binary, but the model provides a continuous predictor

### simulation:
#### how much power do we get from the model in the first place? (sweeping N and m)
#### 

## one way to do it is by adding correlation to x.obs and y that isn't in w.
## in other words, the model is missing an important feature of x.obs that's related to y.
simulate_data <- function(N, m, B0, Bxy, Bzy, seed, prediction_accuracy=0.73, accuracy_imbalance_difference=0.3){
    set.seed(seed)
    # make w and y dependent
    z <- rbinom(N, 1, 0.5)
    x <- rbinom(N, 1, 0.5)

    ystar <- Bzy * z + Bxy * x
    y <- rbinom(N,1,plogis(ystar))

    # glm(y ~ x + z, family="binomial")

    df <- data.table(x=x,y=y,ystar=ystar,z=z)

    if(m < N){
        df <- df[sample(nrow(df), m), y.obs := y]
    } else {
        df <- df[, y.obs := y]
    }

    df <- df[,w_pred:=y]

    pz <- mean(z)

    accuracy_imbalance_ratio <- (prediction_accuracy + accuracy_imbalance_difference/2) / (prediction_accuracy - accuracy_imbalance_difference/2)

    # this works because of conditional probability
    accuracy_z0 <- prediction_accuracy / (pz*(accuracy_imbalance_ratio) + (1-pz))
    accuracy_z1 <- accuracy_imbalance_ratio * accuracy_z0


    yz0 <- df[z==0]$y
    yz1 <- df[z==1]$y
    nz1 <- nrow(df[z==1])
    nz0 <- nrow(df[z==0])

    acc_z0 <- plogis(0.7*scale(yz0) + qlogis(accuracy_z0))
    acc_z1 <- plogis(1.3*scale(yz1) + qlogis(accuracy_z1))
    
    w0z0 <- (1-yz0)**2 + (-1)**(1-yz0) * acc_z0
    w0z1 <- (1-yz1)**2 + (-1)**(1-yz1) * acc_z1
    
    w0z0.noisy.odds <- rlogis(nz0,qlogis(w0z0))
    w0z1.noisy.odds <- rlogis(nz1,qlogis(w0z1))
    df[z==0,w:=plogis(w0z0.noisy.odds)]
    df[z==1,w:=plogis(w0z1.noisy.odds)]

    df[,w_pred:=as.integer(w > 0.5)]

    print(mean(df[y==0]$y == df[y==0]$w_pred))
    print(mean(df[y==1]$y == df[y==1]$w_pred))
    print(mean(df$w_pred == df$y))

    return(df)
}

parser <- arg_parser("Simulate data and fit corrected models")
parser <- add_argument(parser, "--N", default=1000, help="number of observations of w")
parser <- add_argument(parser, "--m", default=500, help="m the number of ground truth observations")
parser <- add_argument(parser, "--seed", default=17, help='seed for the rng')
parser <- add_argument(parser, "--outfile", help='output file', default='example_2.feather')
parser <- add_argument(parser, "--y_explained_variance", help='what proportion of the variance of y can be explained?', default=0.005)
parser <- add_argument(parser, "--prediction_accuracy", help='how accurate is the predictive model?', default=0.73)
parser <- add_argument(parser, "--accuracy_imbalance_difference", help='how much more accurate is the predictive model for one class than the other?', default=0.3)

args <- parse_args(parser)

B0 <- 0
Bxy <- 0.7
Bzy <- -0.7

if(args$m < args$N){
    df <- simulate_data(args$N, args$m, B0, Bxy, Bzy, args$seed, args$prediction_accuracy, args$accuracy_imbalance_difference)

    result <- list('N'=args$N,'m'=args$m,'B0'=B0,'Bxy'=Bxy,'Bzy'=Bzy, 'seed'=args$seed, 'y_explained_variance'=args$y_explained_variance, 'prediction_accuracy'=args$prediction_accuracy, 'accuracy_imbalance_difference'=args$accuracy_imbalance_difference)

    outline <- run_simulation_depvar(df, result, outcome_formula = y ~ x + z, proxy_formula = w_pred ~ y*x + y*z + z*x)

    outfile_lock <- lock(paste0(args$outfile, '_lock'),exclusive=TRUE)

    if(file.exists(args$outfile)){
        logdata <- read_feather(args$outfile)
        logdata <- rbind(logdata,as.data.table(outline),fill=TRUE)
    } else {
        logdata <- as.data.table(outline)
    }

    print(outline)
    write_feather(logdata, args$outfile)
    unlock(outfile_lock)
}
